"evaluate"
Object subclass: #PPParser
instanceVariableNames: 
	' properties '
classVariableNames: ''
poolDictionaries: ''
!
"evaluate" 
PPParser comment: 'An abstract parser for all parsers in PetitParser. Subclasses implement #parseOn: to perform the actual recursive-descent parsing. All parsers support a variety of methods to perform an actual parse, see the methods in the #parsing protocol. Parsers are combined with a series of operators that can be found in the #operations protocol.

Instance Variables:
	properties    <Dictionary>    Stores additional state in the parser object.
'!
"evaluate"  
PPParser subclass: #PPDelegateParser
instanceVariableNames: 
	' parser '
classVariableNames: ''
poolDictionaries: ''
!
"evaluate"   
PPDelegateParser comment: 'A parser that delegates to another parser.

Instance Variables:
	parser	<PPParser>	The parser to delegate to.
'!
"evaluate"  
PPDelegateParser subclass: #PPActionParser
instanceVariableNames: 
	' block '
classVariableNames: ''
poolDictionaries: ''
!
"evaluate"  
PPActionParser comment: 'A parser that performs an action block with the successful parse result of the delegate.

Instance Variables:
	block    <BlockClosure>    The action block to be executed.
'!
"evaluate"
PPActionParser subclass: #PPWrappingParser
instanceVariableNames: ''
classVariableNames: ''
poolDictionaries: ''
!
"evaluate"
PPWrappingParser comment: 'A parser that performs an action block upon activation with the stream and a continuation block.
'!
"evaluate"   
PPDelegateParser subclass: #PPAndParser
instanceVariableNames: ''
classVariableNames: ''
poolDictionaries: ''
!
"evaluate"   
PPAndParser comment: 'The and-predicate, a parser that succeeds whenever its delegate does, but consumes the input stream [Parr 1994, 1995].
'!
"evaluate"  
PPDelegateParser subclass: #PPEndOfInputParser
instanceVariableNames: ''
classVariableNames: ''
poolDictionaries: ''
!
"evaluate"
PPEndOfInputParser comment: 'A parser that succeeds only at the end of the input stream.
'!
"evaluate"  
PPDelegateParser subclass: #PPFlattenParser
instanceVariableNames: ''
classVariableNames: ''
poolDictionaries: ''
!
"evaluate"   
PPFlattenParser comment: 'A parser that answers a flat copy of the range my delegate parses.
'!
"evaluate"  
PPFlattenParser subclass: #PPTokenParser
instanceVariableNames: 
	' tokenClass '
classVariableNames: ''
poolDictionaries: ''
!
"evaluate"   
PPTokenParser comment: 'A parser that answers a token of the range my delegate parses.

Instance Variables:
	tokenClass    <PPToken class>    The token sub-class to be used.
'!
"evaluate"  
PPDelegateParser subclass: #PPMemoizedParser
instanceVariableNames: 
	' stream buffer '
classVariableNames: ''
poolDictionaries: ''
!
"evaluate"
PPMemoizedParser comment: 'a MessageSendNode <a MessageSendNode <a MessageSendNode <a VariableNode<''for''>  a SelectorNode<refraining> >   a SelectorNode<redundant> >   a SelectorNode<computations> > '!
"evaluate"   
PPDelegateParser subclass: #PPNotParser
instanceVariableNames: ''
classVariableNames: ''
poolDictionaries: ''
!
"evaluate"   
PPNotParser comment: 'The not-predicate, a parser that succeeds whenever its delegate does not, but consumes no input [Parr 1994, 1995].
'!
"evaluate"  
PPDelegateParser subclass: #PPOptionalParser
instanceVariableNames: ''
classVariableNames: ''
poolDictionaries: ''
!
"evaluate"  
PPDelegateParser subclass: #PPRepeatingParser
instanceVariableNames: 
	' min max '
classVariableNames: ''
poolDictionaries: ''
!
"evaluate" 
PPRepeatingParser comment: 'A parser that eagerly parses min to max instances of my delegate. The default instance parses eagerly an infinite number of elements, as min is set to 0 and max to infinity (SmallInteger maxVal).

Instance Variables:
	min    <Integer>    The minimum number of repetitions.
	max    <Integer>    The maximum number of repetitions.
'!
"evaluate"  
PPDelegateParser subclass: #PPTrimmingParser
instanceVariableNames: 
	' trimmer '
classVariableNames: ''
poolDictionaries: ''
!
"evaluate"  
PPTrimmingParser comment: 'A parser that silently consumes spaces before and after the delegate parser.
'!
"evaluate"   
PPParser subclass: #PPEpsilonParser
instanceVariableNames: ''
classVariableNames: ''
poolDictionaries: ''
!
"evaluate"   
PPEpsilonParser comment: 'A parser that consumes nothing and always succeeds.
'!
"evaluate" 
PPParser subclass: #PPFailingParser
instanceVariableNames: 
	' message '
classVariableNames: ''
poolDictionaries: ''
!
"evaluate"   
PPFailingParser comment: 'A parser that consumes nothing and always fails.

Instance Variables:
	message <String>	The failure message.
'!
"evaluate"
PPParser subclass: #PPListParser
instanceVariableNames: 
	' parsers '
classVariableNames: ''
poolDictionaries: ''
!
"evaluate"  
PPListParser comment: 'Abstract parser that parses a list of things in some way (to be specified by the subclasses).

Instance Variables:
	parsers    <SequenceableCollection of: PPParser>    A sequence of other parsers to delegate to.
'!
"evaluate" 
PPListParser subclass: #PPChoiceParser
instanceVariableNames: ''
classVariableNames: ''
poolDictionaries: ''
!
"evaluate"
PPChoiceParser comment: 'A parser that uses the first parser that succeeds.
'!
"evaluate"   
PPListParser subclass: #PPSequenceParser
instanceVariableNames: ''
classVariableNames: ''
poolDictionaries: ''
!
"evaluate"  
PPSequenceParser comment: 'A parser that parses a sequence of parsers.
'!
"evaluate"
PPParser subclass: #PPLiteralParser
instanceVariableNames: 
	' literal message '
classVariableNames: ''
poolDictionaries: ''
!
"evaluate"   
PPLiteralParser comment: 'Abstract literal parser that parses some kind of literal type (to be specified by subclasses).

Instance Variables:
	literal    <Object>    The literal object to be parsed.
	message    <String>    The error message to be generated.
'!
"evaluate" 
PPLiteralParser subclass: #PPLiteralObjectParser
instanceVariableNames: ''
classVariableNames: ''
poolDictionaries: ''
!
"evaluate"  
PPLiteralObjectParser comment: 'A parser that accepts a single literal object, such as a character. This is the same as the predicate parser ''PPPredicateParser expect: literal'' but slightly more efficient.
'!
"evaluate"   
PPLiteralParser subclass: #PPLiteralSequenceParser
instanceVariableNames: 
	' size '
classVariableNames: ''
poolDictionaries: ''
!
"evaluate"   
PPLiteralSequenceParser comment: 'a MessageSendNode <a MessageSendNode <a MessageSendNode <a VariableNode<''such''>  a SelectorNode<as> >   a SelectorNode<a> >   a SelectorNode<String> > '!
"evaluate" 
PPParser subclass: #PPPluggableParser
instanceVariableNames: 
	' block '
classVariableNames: ''
poolDictionaries: ''
!
"evaluate"   
PPPluggableParser comment: 'A pluggable parser that passes the parser stream into a block. This enables users to perform manual parsing or to embed other parser frameworks into PetitParser.

Instance Variables:
	block    <BlockClosure>    The pluggable one-argument block.
'!
"evaluate"   
PPParser subclass: #PPPredicateParser
instanceVariableNames: 
	' predicate predicateMessage negated negatedMessage '
classVariableNames: ''
poolDictionaries: ''
!
"evaluate"   
PPPredicateParser comment: 'An abstract parser that accepts if a given predicate holds.

Instance Variables:
	predicate    <BlockClosure>    The block testing for the predicate.
	predicateMessage    <String>    The error message of the predicate.
	negated    <BlockClosure>    The block testing for the negation of the predicate.
	negatedMessage    <String>    The error message of the negated predicate.
'!
"evaluate"
PPPredicateParser subclass: #PPPredicateObjectParser
instanceVariableNames: ''
classVariableNames: ''
poolDictionaries: ''
!
"evaluate"  
PPPredicateObjectParser comment: 'A parser that accepts if a given predicate on one element of the input sequence holds.
'!
"evaluate"  
PPPredicateParser subclass: #PPPredicateSequenceParser
instanceVariableNames: 
	' size '
classVariableNames: ''
poolDictionaries: ''
!
"evaluate"   
PPPredicateSequenceParser comment: 'A parser that accepts if a given predicate on an arbitrary number of elements of the input sequence holds.

Instance Variables:
	size    <Integer>    The number of elements to consume.
'!
!PPPredicateParser methodsFor: #accessing!   
block
	"Answer the predicate block of the receiver."

	^ predicate
!    !
!PPPredicateParser methodsFor: #accessing! 
message
	"Answer the failure message."

	^ predicateMessage
!   !
!PPPredicateParser methodsFor: #printing!  
printNameOn: aStream
	super printNameOn: aStream.
	aStream nextPutAll: ', '; print: predicateMessage
!   !
!PPPredicateObjectParser class methodsFor: #'factory-objects'! 
any
	^ self
		on: [ :each | true ] message: 'input expected'
		negated: [ :each | false ] message: 'no input expected'
!    !
!PPPredicateObjectParser class methodsFor: #'factory-objects'! 
anyExceptAnyOf: aCollectionOfChars
	^ self
		on: [ :each | (aCollectionOfChars includes: each) not ] message: 'any except ' , aCollectionOfChars printString , ' expected'
		negated: [ :each | aCollectionOfChars includes: each ] message: aCollectionOfChars printString ,  ' not expected'
!    !
!PPPredicateObjectParser class methodsFor: #'factory-objects'! 
anyOf: anArray
	^ self
		on: [ :each | anArray includes: each ] message: 'any of ' , anArray printString , ' expected'
		negated: [ :each | (anArray includes: each) not ] message: 'none of ' , anArray printString ,  'expected'
!    !
!PPPredicateObjectParser class methodsFor: #'factory-objects'! 
between: min and: max
	^ self
		on: [ :each | each >= min and: [ each <= max ] ] message: min printString , '..' , max printString , ' expected'
		negated: [ :each | each < min or: [ each > max ] ] message: min printString , '..' , max printString , ' not expected'
! !
!PPPredicateObjectParser class methodsFor: #'factory-objects'! 
expect: anObject
	^ self expect: anObject message: anObject printString , ' expected'
!   !
!PPPredicateObjectParser class methodsFor: #'factory-objects'! 
expect: anObject message: aString
	^ self
		on: [ :each | each = anObject ] message: aString
		negated: [ :each | each ~= anObject ] message: 'no ' , aString
! !
!PPPredicateObjectParser class methodsFor: #'factory-chars'!   
blank
	^ self chars: (String with: Character space with: Character tab) message: 'blank expected'
!   !
!PPPredicateObjectParser class methodsFor: #'factory-chars'!   
char: aCharacter
	^ self expect: aCharacter message: (String with: $" with: aCharacter with: $") , ' expected'
!  !
!PPPredicateObjectParser class methodsFor: #'factory-chars'!   
char: aCharacter message: aString
	^ self expect: aCharacter message: aString
!   !
!PPPredicateObjectParser class methodsFor: #'factory-chars'!   
chars: aCollection message: aString
	^ self on: (PPCharSetPredicate on: [ :char | aCollection includes: char ]) message: aString
!    !
!PPPredicateObjectParser class methodsFor: #'factory-chars'!   
cr
	^ self char: Character cr message: 'carriage return expected'
!   !
!PPPredicateObjectParser class methodsFor: #'factory-chars'!   
digit
	^self
		on: (PPCharSetPredicate on: [:char | char between: $0 and: $9])
		message: 'digit expected'
!    !
!PPPredicateObjectParser class methodsFor: #'factory-chars'!   
hex
	^ self
		on: (PPCharSetPredicate on: [ :char |
			(char between: $0 and: $9)
				or: [ (char between: $a and: $f)
				or: [ (char between: $A and: $F) ] ] ])
		message: 'hex digit expected'
! !
!PPPredicateObjectParser class methodsFor: #'factory-chars'!   
letter
	^self
		on: (PPCharSetPredicate
			on: [:char | (char between: $A and: $Z) or: [char between: $a and: $z]])
		message: 'letter expected'
! !
!PPPredicateObjectParser class methodsFor: #'factory-chars'!   
lf
	^ self char: Character lf
!   !
!PPPredicateObjectParser class methodsFor: #'factory-chars'!   
lowercase
	^ self on: (PPCharSetPredicate on: [ :char | char isLowercase ]) message: 'lowercase letter expected'
!    !
!PPPredicateObjectParser class methodsFor: #'factory-chars'!   
newline
	^ self chars: (String with: Character cr with: Character lf) message: 'newline expected'
!   !
!PPPredicateObjectParser class methodsFor: #'factory-chars'!   
punctuation
	^ self chars: '.,"''?!!;:#$%&()*+-/<>=@[]\^_{}|~' message: 'punctuation expected'
!  !
!PPPredicateObjectParser class methodsFor: #'factory-chars'!   
space
	^ self on: (PPCharSetPredicate on: [ :char | char isSeparator ]) message: 'separator expected'
!   !
!PPPredicateObjectParser class methodsFor: #'factory-chars'!   
tab
	^ self char: Character tab message: 'tab expected'
! !
!PPPredicateObjectParser class methodsFor: #'factory-chars'!   
uppercase
	^ self on: (PPCharSetPredicate on: [ :char | char isUppercase ]) message: 'uppercase letter expected'
!    !
!PPPredicateObjectParser class methodsFor: #'factory-chars'!   
word
	^self
		on: (PPCharSetPredicate
			on: [:char | char isLetter or: [char between: $0 and: $9]])
		message: 'letter or digit expected'
!   !
!PPPredicateObjectParser class methodsFor: #'instance creation'!   
on: aBlock message: aString
	^ self on: aBlock message: aString negated: [ :each | (aBlock value: each) not ] message: 'no ' , aString
!  !
!PPPredicateObjectParser class methodsFor: #'instance creation'!   
on: aBlock message: aString negated: aNegatedBlock message: aNegatedString
	^ self new initializeOn: aBlock message: aString negated: aNegatedBlock message: aNegatedString
! !
!PPPredicateObjectParser methodsFor: #initialization!  
initializeOn: aBlock message: aString negated: aNegatedBlock message: aNegatedString
	predicate := aBlock.
	predicateMessage := aString.
	negated := aNegatedBlock.
	negatedMessage := aNegatedString
!    !
!PPPredicateObjectParser methodsFor: #operators!   
negate
	"Answer a parser that is the negation of the receiving predicate parser."

	^ self class
		on: negated message: negatedMessage
		negated: predicate message: predicateMessage
!   !
!PPPredicateObjectParser methodsFor: #parsing! 
parseOn: aStream
	^ (aStream atEnd not and: [ predicate value: aStream peek ])
		ifFalse: [ PPFailure message: predicateMessage at: aStream position ]
		ifTrue: [ aStream next ]
! !
!PPTokenParser methodsFor: #private!   
create: aCollection start: aStartInteger stop: aStopInteger
	^ self tokenClass on: aCollection start: aStartInteger stop: aStopInteger
!  !
!PPTokenParser methodsFor: #private!   
defaultTokenClass
	^ PPToken
!    !
!PPTokenParser methodsFor: #initialization!
initialize
	tokenClass := self defaultTokenClass
	
! !
!PPTokenParser methodsFor: #accessing! 
tokenClass
	^ tokenClass
!    !
!PPTokenParser methodsFor: #accessing! 
tokenClass: aTokenClass
	tokenClass := aTokenClass
!  !
!PPLiteralObjectParser methodsFor: #parsing!   
parseOn: aStream
	^ (aStream atEnd not and: [ aStream peek = literal ])
		ifFalse: [ PPFailure message: message at: aStream position ]
		ifTrue: [ aStream next ]
! !
!PPLiteralObjectParser methodsFor: #operators! 
caseInsensitive
	"Answer a parser that can parse the receiver case-insensitive."

	literal asUppercase = literal asLowercase ifTrue: [ ^ self ].
	^ PPPredicateObjectParser on: [ :value | literal sameAs: value ] message: message
!  !
!PPChoiceParser methodsFor: #operations!   
/ aRule
	^ self copyWith: aRule
! !
!PPChoiceParser methodsFor: #parsing!  
parseOn: aStream
	"This is optimized code that avoids unnecessary block activations, do not change. When all choices fail, the last failure is answered."

	| element |
	1 to: parsers size do: [ :index |
		element := (parsers at: index)
			parseOn: aStream.
		element isPetitFailure
			ifFalse: [ ^ element ] ].
	^ element
!   !
!PPParser class methodsFor: #'instance creation'!  
named: aString
	^ self new name: aString
!    !
!PPParser class methodsFor: #'instance creation'!  
new
	^ self basicNew initialize
! !
!PPParser methodsFor: #operations! 
, aParser
	"Answer a new parser that parses the receiver followed by aParser."

	^ PPSequenceParser with: self with: aParser
!  !
!PPParser methodsFor: #operations! 
/ aParser
	"Answer a new parser that parses the receiver, if the receiver fails try with aParser (ordered-choice)."

	^ PPChoiceParser with: self with: aParser
!   !
!PPParser methodsFor: #operations! 
| aParser
	"Answer a new parser that either parses the receiver or aParser. Fail if both pass or fail (exclusive choice, unordered choice)."

	^ (self not , aParser) / (aParser not , self) ==> #second
!  !
!PPParser methodsFor: #operations! 
and
	"Answer a new parser (logical and-predicate) that succeeds whenever the receiver does, but never consumes input."

	^ PPAndParser on: self
!   !
!PPParser methodsFor: #operations! 
def: aParser
	"Redefine the receiver as the argument aParser. This method is useful when defining recursive parsers: instantiate a PPParser and later redefine it with another one."

	^ self become: (aParser name: self name)
!   !
!PPParser methodsFor: #operations! 
end
	"Answer a new parser that succeeds at the end of the input and return the result of the receiver."

	^ PPEndOfInputParser on: self
!   !
!PPParser methodsFor: #operations! 
max: anInteger
	"Answer a new parser that parses the receiver at most anInteger times."

	^ PPRepeatingParser on: self max: anInteger
! !
!PPParser methodsFor: #operations! 
memoized
	"Answer a new memoized parser, for refraining redundant computations. This ensures polynomial time O(n^4) for left-recursive grammars and O(n^3) for non left-recursive grammars in the worst case. Not necessary for most grammars that are carefully written and in O(n) anyway."

	^ PPMemoizedParser on: self
!   !
!PPParser methodsFor: #operations! 
min: anInteger
	"Answer a new parser that parses the receiver at least anInteger times."

	^ PPRepeatingParser on: self min: anInteger
!    !
!PPParser methodsFor: #operations! 
min: aMinInteger max: aMaxInteger
	"Answer a new parser that parses the receiver at least aMinInteger and at most aMaxInteger times."

	^ PPRepeatingParser on: self min: aMinInteger max: aMaxInteger
!    !
!PPParser methodsFor: #operations! 
negate
	"Answer a new parser consumes any input token but the receiver."
	
	^ self not , #any asParser ==> #second
!    !
!PPParser methodsFor: #operations! 
not
	"Answer a new parser (logical not-predicate) that succeeds whenever the receiver fails, but never consumes input."

	^ PPNotParser on: self
!  !
!PPParser methodsFor: #operations! 
optional
	"Answer a new parser that parses the receiver, if possible."

	^ PPOptionalParser on: self
!  !
!PPParser methodsFor: #operations! 
plus
	"Answer a new parser that parses the receiver one or more times."

	^ self min: 1
!   !
!PPParser methodsFor: #operations! 
plusGreedy: aParser
	"Answer a new parser that parses the receiver one or more times until it reaches aParser. This is a greedy non-blind implementation of the star operator. aParser is not consumed."

	^ self , (self starGreedy: aParser) map: [ :first :rest | rest copyWithFirst: first ]
!  !
!PPParser methodsFor: #operations! 
plusLazy: aParser
	"Answer a new parser that parses the receiver one or more times until it reaches aParser. This is a lazy non-blind implementation of the star operator. aParser is not consumed."

	^ self , (self starLazy: aParser) map: [ :first :rest | rest copyWithFirst: first ]
!    !
!PPParser methodsFor: #operations! 
star
	"Answer a new parser that parses the receiver zero or more times. This is a greedy and blind implementation that tries to consume as much input as possible and it does not consider what comes afterwards."

	^ PPRepeatingParser on: self
! !
!PPParser methodsFor: #operations! 
starGreedy: aParser
	"Answer a new parser that parses the receiver zero or more times until it reaches aParser. This is a greedy non-blind implementation of the star operator. aParser is not consumed."
	
	| parser |
	parser := PPChoiceParser new.
	parser setParsers: (Array
		with: (self , parser map: [ :each :rest | rest addFirst: each; yourself ])
		with: (aParser and ==> [ :each | OrderedCollection new ])).
	^ parser ==> [ :rest | rest asArray ]
!  !
!PPParser methodsFor: #operations! 
starLazy: aParser
	"Answer a new parser that parses the receiver zero or more times until it reaches aParser. This is a lazy non-blind implementation of the star operator. aParser is not consumed."
	
	| parser |
	parser := PPChoiceParser new.
	parser setParsers: (Array
		with: (aParser and ==> [ :each | OrderedCollection new ])
		with: (self , parser map: [ :each :rest | rest addFirst: each; yourself ])).
	^ parser ==> [ :rest | rest asArray ]
!  !
!PPParser methodsFor: #operations! 
times: anInteger
	"Answer a new parser that parses the receiver exactly anInteger times."

	^ self min: anInteger max: anInteger
!  !
!PPParser methodsFor: #operations! 
wrapped
	"Answer a new parser that is simply wrapped."

	^ PPDelegateParser on: self
!  !
!PPParser methodsFor: #'operations-mapping'!   
==> aBlock
	"Answer a new parser that performs aBlock as action handler on success."

	^ PPActionParser on: self block: aBlock
!    !
!PPParser methodsFor: #'operations-mapping'!   
>=> aBlock
	"Answer a new parser that wraps the receiving parser with a two argument block. The first argument is the parsed stream, the second argument a continuation block on the delegate parser."

	^ PPWrappingParser on: self block: aBlock
!    !
!PPParser methodsFor: #'operations-mapping'!   
answer: anObject
	"Answer a new parser that always returns anObject from a successful parse."

	^ self ==> [ :nodes | anObject ]
!  !
!PPParser methodsFor: #'operations-mapping'!   
flatten
	"Answer a new parser that flattens the underlying collection."

	^ PPFlattenParser on: self
!  !
!PPParser methodsFor: #'operations-mapping'!   
foldLeft: aBlock
	"Answer a new parser that that folds the result of the receiver from left-to-right into aBlock. The argument aBlock must take two or more arguments."
	
	| size args |
	size := aBlock numArgs.
	args := Array new: size.
	^ self ==> [ :nodes |
		args at: 1 put: (nodes at: 1).
		2 to: nodes size by: size - 1 do: [ :index |
			args
				replaceFrom: 2 to: size with: nodes startingAt: index;
				at: 1 put: (aBlock valueWithArguments: args) ].
		args at: 1 ]
! !
!PPParser methodsFor: #'operations-mapping'!   
foldRight: aBlock
	"Answer a new parser that that folds the result of the receiver from right-to-left into aBlock. The argument aBlock must take two or more arguments."

	| size args |
	size := aBlock numArgs.
	args := Array new: size.
	^ self ==> [ :nodes |
		args at: size put: (nodes at: nodes size).
		nodes size - size + 1 to: 1 by: 1 - size do: [ :index |
			args
				replaceFrom: 1 to: size - 1 with: nodes startingAt: index;
				at: size put: (aBlock valueWithArguments: args) ].
		args at: size ]
!    !
!PPParser methodsFor: #'operations-mapping'!   
map: aBlock
	"Answer a new parser that works on the receiving sequence an passes in each element as a block argument."
	
	^ self ==> aBlock
!   !
!PPParser methodsFor: #'operations-mapping'!   
token
	"Answer a new parser that transforms the input to a token."

	^ PPTokenParser on: self
! !
!PPParser methodsFor: #'operations-mapping'!   
token: aTokenClass
	"Answer a new parser that transforms the input to a token of class aTokenClass."

	^ self token tokenClass: aTokenClass
!   !
!PPParser methodsFor: #'operations-mapping'!   
trim
	"Answer a new parser that consumes spaces before and after the receiving parser."

	^ self trimSpaces
!   !
!PPParser methodsFor: #'operations-mapping'!   
trimBlanks
	"Answer a new parser that consumes blanks before and after the receiving parser."

	^ PPTrimmingParser on: self trimmer: #blank asParser
!  !
!PPParser methodsFor: #'operations-mapping'!   
trimSpaces
	"Answer a new parser that consumes spaces before and after the receiving parser."

	^ PPTrimmingParser on: self trimmer: #space asParser
!  !
!PPParser methodsFor: #converting! 
asParser
	^ self
!    !
!PPParser methodsFor: #accessing!  
children
	"Answer a set of child parsers that could follow the receiver."

	^ #()
! !
!PPParser methodsFor: #accessing!  
name
	"Answer the production name of the receiver."

	^ self propertyAt: #name ifAbsent: [ nil ]
!  !
!PPParser methodsFor: #accessing!  
name: aString
	self propertyAt: #name put: aString
!  !
!PPParser methodsFor: #'accessing-properties'! 
hasProperty: aKey
	"Test if the property aKey is present."

	^ properties notNil and: [ properties includesKey: aKey ]
!    !
!PPParser methodsFor: #'accessing-properties'! 
propertyAt: aKey
	"Answer the property value associated with aKey."

	^ self propertyAt: aKey ifAbsent: [ self error: 'Property not found' ]
!  !
!PPParser methodsFor: #'accessing-properties'! 
propertyAt: aKey ifAbsent: aBlock
	"Answer the property value associated with aKey or, if aKey isn't found, answer the result of evaluating aBlock."

	^ properties isNil
		ifTrue: [ aBlock value ]
		ifFalse: [ properties at: aKey ifAbsent: aBlock ]
!    !
!PPParser methodsFor: #'accessing-properties'! 
propertyAt: aKey ifAbsentPut: aBlock
	"Answer the property associated with aKey or, if aKey isn't found store the result of evaluating aBlock as new value."

	^ self propertyAt: aKey ifAbsent: [ self propertyAt: aKey put: aBlock value ]
!  !
!PPParser methodsFor: #'accessing-properties'! 
propertyAt: aKey put: anObject
	"Set the property at aKey to be anObject. If aKey is not found, create a new entry for aKey and set is value to anObject. Answer anObject."

	^ (properties ifNil: [ properties := Dictionary new: 1 ])
		at: aKey put: anObject
! !
!PPParser methodsFor: #'accessing-properties'! 
removeProperty: aKey
	"Remove the property with aKey. Answer the property or raise an error if aKey isn't found."

	^ self removeProperty: aKey ifAbsent: [ self error: 'Property not found' ]
!    !
!PPParser methodsFor: #'accessing-properties'! 
removeProperty: aKey ifAbsent: aBlock
	"
	Remove the property with aKey. Answer the value or, if aKey isn't found, answer
	the result of evaluating aBlock.
	"
	| assoc |
	properties isNil ifTrue: [^aBlock value].
	assoc := properties associationAt: aKey ifAbsent: [^aBlock value].
	properties removeKey: aKey.
	properties isEmpty ifTrue: [properties := nil].
	^assoc value
!   !
!PPParser methodsFor: #initialization! 
initialize
!   !
!PPParser methodsFor: #testing!
isPetitParser
	^ true
!   !
!PPParser methodsFor: #testing!
isUnresolved
	^ false
!   !
!PPParser methodsFor: #parsing!
matches: anObject
	"Answer if anObject can be parsed by the receiver."

	^ (self parse: anObject) isPetitFailure not
!  !
!PPParser methodsFor: #parsing!
matchesIn: anObject
	"Search anObject repeatedly for the matches of the receiver."

	| result |
	result := OrderedCollection new.
	self
		matchesIn: anObject
		do: [ :each | result addLast: each ].
	^ result
!  !
!PPParser methodsFor: #parsing!
matchesIn: anObject do: aBlock
	"Search anObject repeatedly for the matches of the receiver. Evaluate aBlock for each match with the matched parse-tree as the argument. Make sure to always consume exactly one character with each step, to not miss any match."

	((self and ==> aBlock , #any asParser) / #any asParser) star parse: anObject
! !
!PPParser methodsFor: #parsing!
matchingRangesIn: anObject
	"Search anObject repeatedly for the matches of the receiver.  Answer an OrderedCollection of ranges of each match (index of first character to: index of last character)."

	| result |
	result := OrderedCollection new.
	[ :stream | stream position + 1 ] asParser , self , [ :stream | stream position ] asParser
		matchesIn: anObject
		do: [ :value | result addLast: (value first to: value last) ].
	^ result
!   !
!PPParser methodsFor: #parsing!
parse: anObject
	"Parse anObject with the receiving parser and answer the parse-result or an instance of PPFailure."

	^ self parseOn: anObject asPetitStream
! !
!PPParser methodsFor: #parsing!
parse: anObject onError: aBlock
	"Parse anObject with the receiving parser and answer the parse-result or answer the result of evaluating aBlock. Depending on the number of arguments of the block it is simply evaluated, evaluated with the failure object, or evaluated with the error message and position."

	| result |
	result := self parseOn: anObject asPetitStream.
	result isPetitFailure
		ifFalse: [ ^ result ].
	aBlock numArgs = 0
		ifTrue: [ ^ aBlock value ].
	aBlock numArgs = 1
		ifTrue: [ ^ aBlock value: result ].
	^ aBlock value: result message value: result position
!    !
!PPParser methodsFor: #parsing!
parseOn: aStream
	"Parse aStream with the receiving parser and answer the parse-result or an instance of PPFailure. Override this method in subclasses to specify custom parse behavior. Do not call this method from outside, instead use #parse:."

	self subclassResponsibility
!    !
!PPParser methodsFor: #copying!
copy
	^self shallowCopy postCopy
!    !
!PPParser methodsFor: #copying!
postCopy
	properties := properties copy
! !
!PPParser methodsFor: #printing!   
printNameOn: aStream
	self name isNil
		ifTrue: [ aStream print: self hash ]
		ifFalse: [ aStream nextPutAll: self name ]
! !
!PPParser methodsFor: #printing!   
printOn: aStream
	super printOn: aStream.
	aStream nextPut: $(.
	self printNameOn: aStream.
	aStream nextPut: $)
! !
!PPParser methodsFor: #'operations-convenience'!   
delimitedBy: aParser
	"Answer a new parser that parses the receiver one or more times, separated and possibly ended by aParser."
	
	^ (self separatedBy: aParser) , (aParser optional) ==> [ :node |
		node second isNil
			ifTrue: [ node first ]
			ifFalse: [ node first copyWith: node second ] ]
!  !
!PPParser methodsFor: #'operations-convenience'!   
separatedBy: aParser
	"Answer a new parser that parses the receiver one or more times, separated by aParser."
	
	^ (PPSequenceParser with: self with: (PPSequenceParser with: aParser with: self) star) ==> [ :nodes |
		| result |
		result := Array new: 2 * nodes second size + 1.
		result at: 1 put: nodes first.
		nodes second 
			keysAndValuesDo: [ :index :pair | result replaceFrom: 2 * index to: 2 * index + 1 with: pair startingAt: 1 ].
		result ]
!  !
!PPEndOfInputParser methodsFor: #operations!   
end
	^ self
! !
!PPEndOfInputParser methodsFor: #parsing!  
parseOn: aStream
	| position result |
	position := aStream position.
	result := parser parseOn: aStream.
	(result isPetitFailure or: [ aStream atEnd ])
		ifTrue: [ ^ result ].
	result := PPFailure
		message: 'end of input expected'
		at: aStream position.
	aStream position: position.
	^ result
! !
!PPNotParser methodsFor: #parsing! 
parseOn: aStream
	| element position |
	position := aStream position.
	element := parser parseOn: aStream.
	aStream position: position.
	^ element isPetitFailure
		ifFalse: [ PPFailure message: '' at: aStream position ]
!    !
!PPMemoizedParser methodsFor: #operations! 
memoized
	"Ther is no point in memoizing more than once."

	^ self
!    !
!PPMemoizedParser methodsFor: #parsing!
parseOn: aStream
	| memento |
	stream == aStream
		ifFalse: [ self reset: aStream ].
	memento := (buffer at: stream position + 1)
		ifNil: [ buffer at: stream position + 1 put: PPMemento new ].
	memento position isNil
		ifTrue: [
			memento result: (stream size - stream position + 2 < memento count
				ifTrue: [ PPFailure message: 'overflow' at: stream position ]
				ifFalse: [ memento increment. parser parseOn: stream ]).
			memento position: stream position ]
		ifFalse: [ stream position: memento position ].
	^ memento result
!   !
!PPMemoizedParser methodsFor: #private!
reset: aStream
	stream := aStream.
	buffer := Array new: aStream size + 1
!  !
!PPActionParser class methodsFor: #'instance creation'!
on: aParser block: aBlock
	^ (self on: aParser) setBlock: aBlock
!    !
!PPActionParser methodsFor: #accessing!
block
	"Answer the action block of the receiver."

	^ block
!   !
!PPActionParser methodsFor: #parsing!  
parseOn: aStream
	| element |
	^ (element := parser parseOn: aStream) isPetitFailure
		ifFalse: [ block value: element ]
		ifTrue: [ element ]
!   !
!PPActionParser methodsFor: #initialization!   
setBlock: aBlock
	block := aBlock
!   !
!PPOptionalParser methodsFor: #parsing!
parseOn: aStream
	| element |
	element := parser parseOn: aStream.
	^ element isPetitFailure ifFalse: [ ^ element ]
!   !
!PPFlattenParser methodsFor: #hooks!   
create: aCollection start: aStartInteger stop: aStopInteger
	^ aCollection copyFrom: aStartInteger to: aStopInteger
! !
!PPFlattenParser methodsFor: #parsing! 
parseOn: aStream
	| start element stop |
	start := aStream position.
	element := parser parseOn: aStream.
	element isPetitFailure ifTrue: [
		aStream position: start.
		^ element ].
	stop := aStream position.
	^ self create: aStream collection start: start + 1 stop: stop
!  !
!PPListParser class methodsFor: #'instance creation'!  
with: aParser
	^ self withAll: (Array with: aParser)
!    !
!PPListParser class methodsFor: #'instance creation'!  
with: aFirstParser with: aSecondParser
	^ self withAll: (Array with: aFirstParser with: aSecondParser)
!  !
!PPListParser class methodsFor: #'instance creation'!  
withAll: aCollection
	^ self basicNew setParsers: aCollection
!   !
!PPListParser methodsFor: #accessing!  
children
	^ parsers
! !
!PPListParser methodsFor: #copying!
copyWith: aParser
	^ self species withAll: (parsers copyWith: aParser)
!  !
!PPListParser methodsFor: #copying!
postCopy
	super postCopy.
	parsers := parsers copy
! !
!PPListParser methodsFor: #initialization! 
initialize
	super initialize.
	self setParsers: #()
!    !
!PPListParser methodsFor: #initialization! 
setParsers: aCollection
	parsers := aCollection asArray
! !
!PPAndParser methodsFor: #operations!  
and
	^ self
! !
!PPAndParser methodsFor: #parsing! 
parseOn: aStream
	| element position |
	position := aStream position.
	element := parser parseOn: aStream.
	aStream position: position.
	^ element
!  !
!PPEpsilonParser methodsFor: #parsing! 
parseOn: aStream
	^ nil
! !
!PPTrimmingParser class methodsFor: #'instance creation'!  
on: aParser trimmer: aTrimParser
	^ self new
		setParser: aParser;
		setTrimmer: aTrimParser;
		yourself
! !
!PPTrimmingParser methodsFor: #parsing!
parseOn: aStream
	| position element |
	position := aStream position.
	[ (trimmer parseOn: aStream) isPetitFailure ]
		whileFalse.
	element := parser parseOn: aStream.
	element isPetitFailure ifTrue: [
		aStream position: position.
		^ element ].
	[ (trimmer parseOn: aStream) isPetitFailure ]
		whileFalse.
	^ element
!    !
!PPTrimmingParser methodsFor: #initialization! 
setTrimmer: aParser
	trimmer := aParser
! !
!PPTrimmingParser methodsFor: #operations! 
trim
	"There is no point in trimming more than once."

	^ self
!    !
!PPWrappingParser methodsFor: #parsing!
parseOn: aStream
	^ block value: aStream value: [ parser parseOn: aStream ]
! !
!PPPluggableParser class methodsFor: #'instance creation'! 
on: aBlock
	^ self new initializeOn: aBlock
! !
!PPPluggableParser methodsFor: #accessing! 
block
	"Answer the pluggable block."

	^ block
!    !
!PPPluggableParser methodsFor: #initialization!
initializeOn: aBlock
	block := aBlock
!   !
!PPPluggableParser methodsFor: #parsing!   
parseOn: aStream
	| position result |
	position := aStream position.
	result := block value: aStream.
	result isPetitFailure
		ifTrue: [ aStream position: position ].
	^ result
!   !
!PPLiteralParser class methodsFor: #'instance creation'!   
on: anObject
	^ self on: anObject message: anObject printString , ' expected'
!   !
!PPLiteralParser class methodsFor: #'instance creation'!   
on: anObject message: aString
	^ self new initializeOn: anObject message: aString
!   !
!PPLiteralParser methodsFor: #operators!   
caseInsensitive
	"Answer a parser that can parse the receiver case-insensitive."

	self subclassResponsibility
!    !
!PPLiteralParser methodsFor: #initialization!  
initializeOn: anObject message: aString
	literal := anObject.
	message := aString
!  !
!PPLiteralParser methodsFor: #accessing!   
literal
	"Answer the parsed literal."

	^ literal
! !
!PPLiteralParser methodsFor: #accessing!   
message
	"Answer the failure message."

	^ message
!    !
!PPLiteralParser methodsFor: #printing!
printNameOn: aStream
	super printNameOn: aStream.
	aStream nextPutAll: ', '.
	literal storeOn: aStream
!    !
!PPFailingParser class methodsFor: #'instance creation'!   
message: aString
	^ self new setMessage: aString
!    !
!PPFailingParser methodsFor: #accessing!   
message
	"Answer the error message of the receiving parser."

	^ message
!  !
!PPFailingParser methodsFor: #parsing! 
parseOn: aStream
	^ PPFailure message: message at: aStream position
! !
!PPFailingParser methodsFor: #printing!
printNameOn: aStream
	super printNameOn: aStream.
	aStream nextPutAll: ', '; print: message
!    !
!PPFailingParser methodsFor: #initialization!  
setMessage: aString
	message := aString
! !
!PPRepeatingParser class methodsFor: #'instance creation'! 
on: aParser
	^ (super on: aParser) setMin: 0 max: SmallInteger maxVal
!   !
!PPRepeatingParser class methodsFor: #'instance creation'! 
on: aParser max: aMaxInteger
	^ (self on: aParser) setMin: 0 max: aMaxInteger
!   !
!PPRepeatingParser class methodsFor: #'instance creation'! 
on: aParser min: aMinInteger
	^ (self on: aParser) setMin: aMinInteger max: SmallInteger maxVal 
!    !
!PPRepeatingParser class methodsFor: #'instance creation'! 
on: aParser min: aMinInteger max: aMaxInteger
	^ (self on: aParser) setMin: aMinInteger max: aMaxInteger
!    !
!PPRepeatingParser methodsFor: #accessing! 
max
	"Answer the maximum number of repetitions."

	^ max
!  !
!PPRepeatingParser methodsFor: #accessing! 
min
	"Answer the minimum number of repetitions."

	^ min
!  !
!PPRepeatingParser methodsFor: #parsing!   
parseOn: aStream
	| start element elements |
	start := aStream position.
	elements := OrderedCollection new.
	[ elements size < min ] whileTrue: [
		(element := parser parseOn: aStream) isPetitFailure ifTrue: [
			aStream position: start.
			^ element ].
		elements addLast: element ].
	[ elements size < max ] whileTrue: [
		(element := parser parseOn: aStream) isPetitFailure
			ifTrue: [ ^ elements asArray ].
		elements addLast: element ].
	^ elements asArray
! !
!PPRepeatingParser methodsFor: #printing!  
printOn: aStream
	super printOn: aStream.
	aStream nextPutAll: ' ['; print: min; nextPutAll: ', '; nextPutAll: (max = SmallInteger maxVal
		ifTrue: [ '*' ] ifFalse: [ max printString ]); nextPut: $]
!    !
!PPRepeatingParser methodsFor: #initialization!
setMin: aMinInteger max: aMaxInteger
	min := aMinInteger.
	max := aMaxInteger
!  !
!PPSequenceParser methodsFor: #operations! 
, aRule
	^ self copyWith: aRule
! !
!PPSequenceParser methodsFor: #operations! 
map: aBlock
	^ self ==> [ :nodes | aBlock valueWithArguments: nodes ]
!   !
!PPSequenceParser methodsFor: #parsing!
parseOn: aStream
	"This is optimized code that avoids unnecessary block activations, do not change."

	| start elements element |
	start := aStream position.
	elements := Array new: parsers size.
	1 to: parsers size do: [ :index |
		element := (parsers at: index)
			parseOn: aStream.
		element isPetitFailure ifTrue: [
			aStream position: start.
			^ element ].
		elements at: index put: element ].
	^ elements
!    !
!PPSequenceParser methodsFor: #permutation!
permutation: anArrayOfIntegers
	"Answer a permutation of the receivers sequence."
	
	anArrayOfIntegers do: [ :index |
		(index isInteger and: [ index between: 1 and: parsers size ])
			ifFalse: [ self error: 'Invalid permutation index: ' , index printString ] ].
	^ self ==> [ :nodes | anArrayOfIntegers collect: [ :index | nodes at: index ] ]
!    !
!PPDelegateParser class methodsFor: #'instance creation'!  
on: aParser
	^ self new setParser: aParser
!  !
!PPDelegateParser methodsFor: #accessing!  
children
	^ Array with: parser
!  !
!PPDelegateParser methodsFor: #parsing!
parseOn: aStream
	^ parser parseOn: aStream
! !
!PPDelegateParser methodsFor: #initialization! 
setParser: aParser
	parser := aParser
!   !
!PPPredicateSequenceParser class methodsFor: #'instance creation'! 
on: aBlock message: aString negated: aNegatedBlock message: aNegatedString size: anInteger
	^ self new initializeOn: aBlock message: aString negated: aNegatedBlock message: aNegatedString size: anInteger
! !
!PPPredicateSequenceParser class methodsFor: #'instance creation'! 
on: aBlock message: aString size: anInteger
	^ self on: aBlock message: aString negated: [ :each | (aBlock value: each) not ] message: 'no ' , aString size: anInteger 
! !
!PPPredicateSequenceParser methodsFor: #initialization!
initializeOn: aBlock message: aString negated: aNegatedBlock message: aNegatedString size: anInteger
	predicate := aBlock.
	predicateMessage := aString.
	negated := aNegatedBlock.
	negatedMessage := aNegatedString.
	size := anInteger 
!  !
!PPPredicateSequenceParser methodsFor: #operators! 
negate
	"Answer a parser that is the negation of the receiving predicate parser."

	^ self class
		on: negated message: negatedMessage
		negated: predicate message: predicateMessage
		size: size
! !
!PPPredicateSequenceParser methodsFor: #parsing!   
parseOn: aStream
	| position result |
	position := aStream position.
	result := aStream next: size.
	(result size = size and: [ predicate value: result ])
		ifTrue: [ ^ result ].
	aStream position: position.
	^ PPFailure message: predicateMessage at: aStream position
!   !
!PPPredicateSequenceParser methodsFor: #accessing! 
size
	"Answer the sequence size of the receiver."

	^ size
!    !
!PPLiteralSequenceParser methodsFor: #initialization!  
initializeOn: anObject message: aString
	super initializeOn: anObject message: aString.
	size := literal size
!  !
!PPLiteralSequenceParser methodsFor: #parsing! 
parseOn: aStream
	| position result |
	position := aStream position.
	result := aStream next: size.
	result = literal ifTrue: [ ^ result ].
	aStream position: position.
	^ PPFailure message: message at: aStream position
!    !
!PPLiteralSequenceParser methodsFor: #accessing!   
size
	"Answer the sequence size of the receiver."

	^ size
!    !
!PPLiteralSequenceParser methodsFor: #operators!   
caseInsensitive
	"Answer a parser that can parse the receiver case-insensitive."

	literal asUppercase = literal asLowercase ifTrue: [ ^ self ].
	^ PPPredicateSequenceParser on: [ :value | literal sameAs: value ] message: message size: size
! !
!Set methods!  
asParser
	^ PPChoiceParser withAll: (self collect: [ :each | each asParser ])
!   !
!String methods!   
asParser
	^ PPLiteralSequenceParser on: self
!    !
!Symbol methods!   
asParser
	^ PPPredicateObjectParser perform: self
!   !
!UndefinedObject methods!  
asParser
	^ PPEpsilonParser new
! !
